"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeliLambdaFactory = void 0;
const events_1 = require("events");
const util_1 = require("util");
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const lambda_1 = require("./lambda");
const checkpointManager_1 = require("./checkpointManager");
const getDefaultCheckpoint = () => {
    return {
        clients: undefined,
        durableSequenceNumber: 0,
        expHash1: server_services_client_1.defaultHash,
        logOffset: -1,
        sequenceNumber: 0,
        signalClientConnectionNumber: 0,
        lastSentMSN: 0,
        nackMessages: undefined,
        successfullyStartedLambdas: [],
        checkpointTimestamp: Date.now(),
    };
};
class DeliLambdaFactory extends events_1.EventEmitter {
    constructor(operationsDbMongoManager, documentRepository, checkpointService, tenantManager, clientManager, forwardProducer, signalProducer, reverseProducer, serviceConfiguration, restartOnCheckpointFailure, kafkaCheckpointOnReprocessingOp) {
        super();
        this.operationsDbMongoManager = operationsDbMongoManager;
        this.documentRepository = documentRepository;
        this.checkpointService = checkpointService;
        this.tenantManager = tenantManager;
        this.clientManager = clientManager;
        this.forwardProducer = forwardProducer;
        this.signalProducer = signalProducer;
        this.reverseProducer = reverseProducer;
        this.serviceConfiguration = serviceConfiguration;
        this.restartOnCheckpointFailure = restartOnCheckpointFailure;
        this.kafkaCheckpointOnReprocessingOp = kafkaCheckpointOnReprocessingOp;
    }
    async create(config, context) {
        const { documentId, tenantId } = config;
        const sessionMetric = (0, utils_1.createSessionMetric)(tenantId, documentId, server_services_telemetry_1.LumberEventName.SessionResult, this.serviceConfiguration);
        const sessionStartMetric = (0, utils_1.createSessionMetric)(tenantId, documentId, server_services_telemetry_1.LumberEventName.StartSessionResult, this.serviceConfiguration);
        const messageMetaData = {
            documentId,
            tenantId,
        };
        let gitManager;
        let document;
        try {
            // Lookup the last sequence number stored
            // TODO - is this storage specific to the orderer in place? Or can I generalize the output context?
            document = await this.documentRepository.readOne({ documentId, tenantId });
            // Check if the document was deleted prior.
            if (!(0, utils_1.isDocumentValid)(document)) {
                // (Old, from tanviraumi:) Temporary guard against failure until we figure out what causing this to trigger.
                // Document sessions can be joined (via Alfred) after a document is functionally deleted.
                const errorMessage = `Received attempt to connect to a missing/deleted document.`;
                context.log?.error(errorMessage, { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(errorMessage, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                return new utils_1.NoOpLambda(context);
            }
            if (!(0, utils_1.isDocumentSessionValid)(document, this.serviceConfiguration)) {
                // Session for this document is either nonexistent or exists in a different location.
                const errMsg = `Received attempt to connect to invalid session: ${JSON.stringify(document.session)}`;
                context.log?.error(errMsg, { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(errMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                if (this.serviceConfiguration.enforceDiscoveryFlow) {
                    // This can/will prevent any users from creating a valid session in this location
                    // for the liftime of this NoOpLambda. This is not ideal; however, throwing an error
                    // to prevent lambda creation would mark the document as corrupted, which is worse.
                    return new utils_1.NoOpLambda(context);
                }
            }
            gitManager = await this.tenantManager.getTenantGitManager(tenantId, documentId);
        }
        catch (error) {
            const errMsg = "Deli lambda creation failed";
            context.log?.error(`${errMsg}. Exception: ${(0, util_1.inspect)(error)}`, { messageMetaData });
            server_services_telemetry_1.Lumberjack.error(errMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId), error);
            this.logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg);
            throw error;
        }
        let lastCheckpoint;
        // Restore deli state if not present in the cache. Mongodb casts undefined as null so we are checking
        // both to be safe. Empty sring denotes a cache that was cleared due to a service summary or the document
        // was created within a different tenant.
        if (document.deli === undefined || document.deli === null) {
            const message = "New document. Setting empty deli checkpoint";
            context.log?.info(message, { messageMetaData });
            server_services_telemetry_1.Lumberjack.info(message, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
            lastCheckpoint = getDefaultCheckpoint();
        }
        else {
            if (document.deli === "") {
                const docExistsMessge = "Existing document. Fetching checkpoint from summary";
                context.log?.info(docExistsMessge, { messageMetaData });
                server_services_telemetry_1.Lumberjack.info(docExistsMessge, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                const lastCheckpointFromSummary = await this.loadStateFromSummary(tenantId, documentId, gitManager, context.log);
                if (lastCheckpointFromSummary === undefined) {
                    const errMsg = "Could not load state from summary";
                    context.log?.error(errMsg, { messageMetaData });
                    server_services_telemetry_1.Lumberjack.error(errMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                    this.logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg);
                    lastCheckpoint = getDefaultCheckpoint();
                }
                else {
                    lastCheckpoint = lastCheckpointFromSummary;
                    // Since the document was originated elsewhere or cache was cleared, logOffset info is irrelavant.
                    // Currently the lambda checkpoints only after updating the logOffset so setting this to lower
                    // is okay. Conceptually this is similar to default checkpoint where logOffset is -1. In this case,
                    // the sequence number is 'n' rather than '0'.
                    lastCheckpoint.logOffset = -1;
                    const message = `Deli checkpoint from summary: ${JSON.stringify(lastCheckpoint)}`;
                    context.log?.info(message, { messageMetaData });
                    server_services_telemetry_1.Lumberjack.info(message, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                }
            }
            else {
                lastCheckpoint = await this.checkpointService.restoreFromCheckpoint(documentId, tenantId, "deli", document);
            }
        }
        // Add checkpointTimestamp as UTC now if checkpoint doesn't have a timestamp yet.
        if (lastCheckpoint.checkpointTimestamp === undefined ||
            lastCheckpoint.checkpointTimestamp === null) {
            lastCheckpoint.checkpointTimestamp = Date.now();
        }
        const checkpointManager = (0, checkpointManager_1.createDeliCheckpointManagerFromCollection)(tenantId, documentId, this.checkpointService);
        const deliLambda = new lambda_1.DeliLambda(context, tenantId, documentId, lastCheckpoint, checkpointManager, this.clientManager, 
        // The producer as well it shouldn't take. Maybe it just gives an output stream?
        this.forwardProducer, this.signalProducer, this.reverseProducer, this.serviceConfiguration, sessionMetric, sessionStartMetric, this.checkpointService, this.restartOnCheckpointFailure, this.kafkaCheckpointOnReprocessingOp);
        deliLambda.on("close", (closeType) => {
            const handler = async () => {
                if (closeType === server_services_core_1.LambdaCloseType.ActivityTimeout ||
                    closeType === server_services_core_1.LambdaCloseType.Error) {
                    if (document?.isEphemeralContainer) {
                        // Call to historian to delete summaries
                        await (0, server_services_core_1.requestWithRetry)(async () => gitManager.deleteSummary(false), "deliLambda_onClose" /* callName */, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId) /* telemetryProperties */, (error) => true /* shouldRetry */, 3 /* maxRetries */);
                        // Delete the document metadata
                        await (0, server_services_core_1.runWithRetry)(async () => this.documentRepository.deleteOne({
                            documentId,
                            tenantId,
                        }), "deleteDocMetadata", 3 /* maxRetries */, 1000 /* retryAfterMs */, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId), (error) => error.code === 11000 /* shouldIgnoreError */, (error) => true /* shouldRetry */);
                        server_services_telemetry_1.Lumberjack.info(`Successfully cleaned up ephemeral container`, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                        return;
                    }
                    const filter = { documentId, tenantId, session: { $exists: true } };
                    const data = {
                        "session.isSessionAlive": false,
                        "session.isSessionActive": false,
                        "lastAccessTime": Date.now(),
                    };
                    await this.documentRepository.updateOne(filter, data, undefined);
                    const message = `Marked session alive and active as false for closeType:
                        ${JSON.stringify(closeType)}`;
                    context.log?.info(message, { messageMetaData });
                    server_services_telemetry_1.Lumberjack.info(message, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                }
            };
            handler().catch((e) => {
                const message = `Failed to handle session alive and active with exception ${e}`;
                context.log?.error(message, { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(message, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId), e);
            });
        });
        // Fire-and-forget sessionAlive and sessionActive update for session-boot performance.
        // Worst case is that document is allowed to be deleted while active.
        context.log?.info(`Deli Lambda is marking session as alive and active as true.`, {
            messageMetaData,
        });
        this.documentRepository
            .updateOne({ tenantId, documentId }, {
            "session.isSessionAlive": true,
            "session.isSessionActive": true,
        })
            .catch((error) => {
            const errMsg = "Deli Lambda failed to mark session as active.";
            context.log?.error(`${errMsg} Exception: ${(0, util_1.inspect)(error)}`, { messageMetaData });
            server_services_telemetry_1.Lumberjack.error(`${errMsg}`, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId), error);
        });
        return deliLambda;
    }
    logSessionFailureMetrics(sessionMetric, sessionStartMetric, errMsg) {
        sessionMetric?.error(errMsg);
        sessionStartMetric?.error(errMsg);
    }
    async dispose() {
        const mongoClosedP = this.operationsDbMongoManager.close();
        const forwardProducerClosedP = this.forwardProducer.close();
        const signalProducerClosedP = this.signalProducer?.close();
        const reverseProducerClosedP = this.reverseProducer.close();
        await Promise.all([
            mongoClosedP,
            forwardProducerClosedP,
            signalProducerClosedP,
            reverseProducerClosedP,
        ]);
    }
    // Fetches last durable deli state from summary. Returns undefined if not present.
    async loadStateFromSummary(tenantId, documentId, gitManager, logger) {
        const existingRef = await gitManager.getRef(encodeURIComponent(documentId));
        if (existingRef) {
            try {
                const content = await gitManager.getContent(existingRef.object.sha, ".serviceProtocol/deli");
                const summaryCheckpoint = JSON.parse((0, common_utils_1.toUtf8)(content.content, content.encoding));
                return summaryCheckpoint;
            }
            catch (exception) {
                const messageMetaData = {
                    documentId,
                    tenantId,
                };
                const errorMessage = `Error fetching deli state from summary`;
                logger?.error(errorMessage, { messageMetaData });
                logger?.error(JSON.stringify(exception), { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(errorMessage, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId), exception);
                return undefined;
            }
        }
    }
}
exports.DeliLambdaFactory = DeliLambdaFactory;
//# sourceMappingURL=lambdaFactory.js.map