"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CheckpointManager = void 0;
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * MongoDB specific implementation of ICheckpointManager
 */
class CheckpointManager {
    constructor(context, tenantId, documentId, documentRepository, opCollection, deltaService, getDeltasViaAlfred, verifyLastOpPersistence, checkpointService) {
        this.context = context;
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.documentRepository = documentRepository;
        this.opCollection = opCollection;
        this.deltaService = deltaService;
        this.getDeltasViaAlfred = getDeltasViaAlfred;
        this.verifyLastOpPersistence = verifyLastOpPersistence;
        this.checkpointService = checkpointService;
        this.clientFacadeRetryEnabled = (0, server_services_core_1.isRetryEnabled)(this.opCollection);
    }
    /**
     * Writes the checkpoint information to MongoDB
     */
    async write(checkpoint, protocolHead, pending, noActiveClients, globalCheckpointOnly, markAsCorrupt = false) {
        const isLocalCheckpoint = !noActiveClients && !globalCheckpointOnly;
        if (this.getDeltasViaAlfred) {
            if (pending.length > 0 && this.verifyLastOpPersistence) {
                // Verify that the last pending op has been persisted to op storage
                // If it is, we can checkpoint
                const expectedSequenceNumber = pending[pending.length - 1].operation.sequenceNumber;
                const lastDelta = await this.deltaService.getDeltas("", this.tenantId, this.documentId, expectedSequenceNumber - 1, expectedSequenceNumber + 1, "scribe");
                // If we don't get the expected delta, retry after a delay
                if (lastDelta.length === 0 ||
                    lastDelta[0].sequenceNumber < expectedSequenceNumber) {
                    const lumberjackProperties = {
                        ...(0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId),
                        expectedSequenceNumber,
                        lastDelta: lastDelta.length > 0 ? lastDelta[0].sequenceNumber : -1,
                    };
                    server_services_telemetry_1.Lumberjack.info(`Pending ops were not been persisted to op storage. Retrying after delay`, lumberjackProperties);
                    await (0, common_utils_1.delay)(1500);
                    const lastDelta1 = await this.deltaService.getDeltas("", this.tenantId, this.documentId, expectedSequenceNumber - 1, expectedSequenceNumber + 1, "scribe");
                    if (lastDelta1.length === 0 ||
                        lastDelta1[0].sequenceNumber < expectedSequenceNumber) {
                        const errMsg = "Pending ops were not been persisted to op storage. Checkpointing failed";
                        server_services_telemetry_1.Lumberjack.error(errMsg, lumberjackProperties);
                        throw new Error(errMsg);
                    }
                    server_services_telemetry_1.Lumberjack.info(`Verified on retry that pending ops are persisted`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                }
            }
            await this.checkpointService.writeCheckpoint(this.documentId, this.tenantId, "scribe", checkpoint, isLocalCheckpoint, markAsCorrupt);
        }
        else {
            // The order of the three operations below is important.
            // We start by writing out all pending messages to the database. This may be more messages that we would
            // have seen at the current checkpoint we are trying to write (because we continue process messages while
            // waiting to write a checkpoint) but is more efficient and simplifies the code path.
            //
            // We then write the update to the document collection. This marks a log offset inside of MongoDB at which
            // point if Kafka restartes we will not do work prior to this logOffset. At this point the snapshot
            // history has been written, all ops needed are written, and so we can store the final mark.
            //
            // And last we delete all mesages in the list prior to the summaryprotocol sequence number. From now on these
            // will no longer be referenced.
            const dbOps = pending.map((message) => ({
                ...message,
                mongoTimestamp: new Date(message.operation.timestamp),
            }));
            if (dbOps.length > 0) {
                await (0, server_services_core_1.runWithRetry)(async () => this.opCollection.insertMany(dbOps, false), "writeCheckpointScribe", 3 /* maxRetries */, 1000 /* retryAfterMs */, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), (error) => error.code === 11000 /* shouldIgnoreError */, (error) => !this.clientFacadeRetryEnabled /* shouldRetry */);
            }
            // Write out the full state first that we require to global & local DB
            await this.checkpointService.writeCheckpoint(this.documentId, this.tenantId, "scribe", checkpoint, isLocalCheckpoint, markAsCorrupt);
            // And then delete messagses that were already summarized.
            await this.opCollection.deleteMany({
                "documentId": this.documentId,
                "operation.sequenceNumber": { $lte: protocolHead },
                "tenantId": this.tenantId,
            });
        }
    }
    /**
     * Removes the checkpoint information from MongoDB
     */
    async delete(sequenceNumber, lte) {
        // Clears the checkpoint information from mongodb.
        await this.documentRepository.updateOne({
            documentId: this.documentId,
            tenantId: this.tenantId,
        }, {
            scribe: "",
        }, null);
        // And then delete messagse we no longer will reference
        await this.opCollection.deleteMany({
            "documentId": this.documentId,
            "operation.sequenceNumber": lte ? { $lte: sequenceNumber } : { $gte: sequenceNumber },
            "tenantId": this.tenantId,
        });
    }
}
exports.CheckpointManager = CheckpointManager;
//# sourceMappingURL=checkpointManager.js.map